/***************************************************************************/
/* ALGORITHME Essai/Erreur                                                                  */
/***************************************************************************/

// ===== Includes =====

// ----- Librairies standard -----
#include <stdlib.h>
#include <stdio.h>

// ----- Sources du projet -------

#include "debug.h"

#include "sudoku.h"
#include "algorithme.h"

#include "sudoku_funcs.h"
#include "algorithme_funcs.h"

#include "valid.h"
#include "ihm.h"

// ===== Types Locaux =====

// ===== Prototypes Locaux =====

// ===== Fonctions ======

// ----- Gestion de la liste des coups -----

//Initialise une liste de coups vide.
void ALG_f_init_list(ALG_list_coups_t *pL)
{
  pL->nb_coups_max = 0;
  pL->nb_coups = 0;
  pL->idx_coup = 0;      // Index du prochain de ces coups  jouer
  pL->coups    = NULL;   // Aucun coup dans la liste.
  JEU_f_init_explain(&pL->explain); // Pour que l'algorithme puisse commenter ses choix
}

//Ajoute une copie de *pc  la liste (on nglige les problmes de mmoire).
void ALG_f_add_list(ALG_list_coups_t *pL, JEU_coup_t *pc)
{
  int size, new_size;

  //Rservation de mmoire
  size = pL->nb_coups;
  if (size == pL->nb_coups_max)
  {
    if (size == 0)
      new_size = 1;
    else
      new_size = 2*size;
    pL->coups = (JEU_coup_t*) m_dbg_realloc(pL->coups, new_size*sizeof(JEU_coup_t));
    pL->nb_coups_max = new_size;
  }

  //Copie de l'lment  ajouter.
  pL->coups[size] = *pc; //Copie de structure
  pL->nb_coups    = size+1;
}

//Vide la liste
void ALG_f_clear_list(ALG_list_coups_t *pL)
{
  int i;
  for(i=0; i<pL->nb_coups; i++)
  {
    JEU_f_clear_coup(&pL->coups[i]);
  }
  pL->nb_coups = 0;
  pL->idx_coup = 0;      // Index du prochain de ces coups  jouer
}

// ----- Gestion du plan d'exploration -----

//Initialise le plan d'exploration
void ALG_f_explore_init(ALG_plan_explore_t *pe)
{
  int i;
  int size;
  ALG_list_coups_t *pcL;

  //size = JEU_f_get_explore_size(); //Nombre d'tapes maximal  prvoir
  size = 0;

  pe->coups_cases = (ALG_list_coups_t *) m_dbg_malloc(size*sizeof(ALG_list_coups_t));
  pe->a_idx_coups = (int*) m_dbg_malloc(size*sizeof(int));
  pe->nb_max = size;

  //Pour chaque tape ...
  for(i=0; i<size; i++)
  {
    pcL = &pe->coups_cases[i];
    ALG_f_init_list(pcL);     //Prparer la mmoire pour une liste de coups.
    pe->a_idx_coups[i] = (-1);    //Aucun coup n'a t dtermin pour cette tape
  }
}

//S'assure qu'on a de la mmoire pour l'tape d'indice 'i_explo'
void ALG_f_explore_check(ALG_plan_explore_t *pe, int i_explo)
{
  if (i_explo >= pe->nb_max)
  {
    int nb;
    int i;
    ALG_list_coups_t *pcL;

    if (pe->nb_max > 0) //Quelle taille allouer?
      nb = 2 * pe->nb_max;
    else
      nb = 1;

    //Allouer...
    pe->coups_cases = (ALG_list_coups_t *) m_dbg_realloc(pe->coups_cases, nb*sizeof(ALG_list_coups_t));
    pe->a_idx_coups = (int*) m_dbg_realloc(pe->a_idx_coups, nb*sizeof(int));

    //Initialiser les nouvelles tapes
    for(i=pe->nb_max; i<nb; i++)
    {
      pcL = &pe->coups_cases[i];
      ALG_f_init_list(pcL);       //Prparer la mmoire pour une liste de coups.
      pe->a_idx_coups[i] = (-1);  //Aucun coup n'a t dtermin pour cette tape
    }

    pe->nb_max = nb; //Nouvelle taille
  }
}

//Libre la mmoire qui avait t alloue dans ce plan d'exploration
void ALG_f_explore_reset(ALG_plan_explore_t *pe)
{
  int i;
  ALG_list_coups_t *pcL;

  //Pour chaque tape ...
  for(i=0; i<pe->nb_max; i++)
  {
    pcL = &pe->coups_cases[i];
    ALG_f_clear_list(pcL);     //Librer la mmoire rserve pour cette tape.
  }

  m_dbg_free(pe->coups_cases); //Puis librer les tableaux.
  m_dbg_free(pe->a_idx_coups);

  pe->nb_max = 0;              //Puis marquer qu'on a tout libr.
  pe->a_idx_coups = NULL;
  pe->coups_cases = NULL;
}

// ----- Rsolution du jeu par un algorithme essai/erreur -----
#define DEBUG_ON 0

//Arguments:
//  p_position: dcrit la position de dpart
//Retour:
//  TRUE: recherche excute sans problme (qu'on ait trouv ou non des solutions)
//  FALSE: problme rencontr.
bool_t ALG_f_resoudre(JEU_position_t *p_position)
{
  int i_explo;
  int i_coup;  //Index du coup  jouer dans l'tape courante
  ALG_plan_explore_t exploration;
  ALG_list_coups_t *pcL; //Liste de coups pour l'tape courante.
  JEU_coup_t *pc;
  int nb_loops = 0; //DEBUG
  int nb_solutions = 0; //NOMBRE DES SOLUTIONS TROUVEES
  bool_t r;

  //Position initiale
  r = JEU_f_init(p_position);
  if (r == FALSE)
  {
    ihm_out("ERREUR: Problme d'initialisation\n");
    return FALSE;
  }

  ALG_f_explore_init(&exploration);

  //Tant que i_explo>=0
  i_explo = 0;
  while(i_explo >= 0)
  {
    nb_loops++;

    //Gestion de la mmoire (en cas de besoin)
    ALG_f_explore_check(&exploration, i_explo);

    //Pointer l'tape courante
    pcL = &exploration.coups_cases[i_explo];

    //La liste des coups reste-t-elle  tablir?
    if(exploration.a_idx_coups[i_explo] == -1)
    {
      //IFDBG { ihm_out("==>"); ihm_out("i_explo=%d, i_coup=%d\n", i_explo, i_coup); }
      JEU_f_lister_coups(pcL);
      i_coup = 0;
      exploration.a_idx_coups[i_explo] = 0; //Maintenant c'est fait ...

      //Est-ce que l'on a une solution?
      if ( TRUE == JEU_f_est_solution())
      {
        JEU_position_t *ppos;

        nb_solutions++;
        //Affichage de la solution.
        ihm_out("\nSOLUTION %d:\n", nb_solutions);
        ppos = JEU_f_get_position();
        ihm_affiche(ppos);
      }
    }
    //Sinon c'est qu'il faut annuler le dernier coup jou.
    else
    {
	     i_coup = pcL->idx_coup;
       //IFDBG { ihm_out("<=="); ihm_out("i_explo=%d, i_coup=%d\n", i_explo, i_coup); }
       pc = &pcL->coups[i_coup];
       JEU_f_annuler(pc);
       i_coup++;
    }

    //DEBUG: voir s'il y a quelque-chose de louche ...
    IFDBG { ihm_out("\n"); } //JEU_f_print_board_n(4); }
    IFDBG { JEU_f_print_list(3,7); }
    IFDBG { JEU_position_t *ppos; ppos = JEU_f_get_position(); ihm_affiche(ppos); }

    //  S'il reste des coups  jouer
    if (i_coup < pcL->nb_coups)
    {
      //Essayer le suivant
      IFDBG { ihm_out("==>Jouer i_coup=%d/%d\n", i_coup, pcL->nb_coups);  }
      pcL->idx_coup = i_coup;
      pc = &pcL->coups[i_coup];
      if (IHM_v_opt_expliquer)
        JEU_f_print_expliquer(i_explo, &pcL->explain, pc);
      JEU_f_jouer(pc);
      i_explo++;
    }
    //  Sinon
    else
    {
      //Revenir sur le dernier coup jou
      exploration.a_idx_coups[i_explo] = -1; //La liste de coups sera  refaire...
      IFDBG { JEU_f_print_board(); }
      i_explo--;
    }
  }  //while (i_explo >= 0)

  //Afficher le bilan ...
  ihm_out("\nBILAN: %d SOLUTIONS en %d BOUCLES!\n", nb_solutions, nb_loops);

  //Librer la mmoire du module 'jeu'
  JEU_f_reset();
  //Librer la mmoire du module "algorithme".
  ALG_f_explore_reset(&exploration);

  //Fin
  return TRUE;
}

// ===== FIN DE FICHIER =====
