// =======================================
// Routines de debug gnralistes commodes
// =======================================

// ===== Includes =====

// ----- Librairies standard -----
#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>

// ----- Sources du projet -------

#include "debug.h"

// ===== MACROS LOCALES =====

#define M_DBG_ALLOC_HEADER_PREFIX 0x12482359
#define M_DBG_ALLOC_HEADER_SIZE   offsetof(t_dbg_alloc_block,contents)
#define M_DBG_FREE_HEADER_PREFIX  0x95328421

// ===== Types locaux =====

typedef struct {
  unsigned int prefix;
  unsigned int size;
  char contents;
} t_dbg_alloc_block;

// ===== Variables =====

#ifdef DEBUG_MEM

static int v_dbg_alloc_bytes;
static int v_dbg_alloc_count;
static int v_dbg_realloc_count;
static int v_dbg_free_count;
static int v_dbg_free_bytes;
static int v_dbg_mem_on =0;
static int v_dbg_count_max;
static int v_dbg_bytes_max;

#endif //#ifdef DEBUG_MEM

// ===== Fonctions =====

static void f_dbg_mult_free_aux(void *p, int dim, va_list ap);
static void *f_dbg_mult_alloc_aux(int dim, int es, va_list ap);
static void f_dbg_mult_memcpy_aux(void *tgt, void *src, int dim, int es, va_list ap);
static int f_dbg_mult_memcmp_aux(void *a0, void *a1, int dim, int es, va_list ap);

// ----- Dbogage mmoire -----

#ifdef DEBUG_MEM

//Reset des compteurs d'allocation
void f_dbg_start_alloc_count(void)
{
  if (v_dbg_mem_on != 0)
    printf("WARNING! f_dbg_reset_alloc_count dj appel!!!\n");
  v_dbg_alloc_bytes   = 0;
  v_dbg_alloc_count   = 0;
  v_dbg_realloc_count = 0;
  v_dbg_free_count    = 0;
  v_dbg_free_bytes    = 0;
  v_dbg_count_max     = 0;
  v_dbg_bytes_max     = 0;
  v_dbg_mem_on        = 1;
}

//Affiche les compteurs d'allocation actuels
void f_dbg_print_counters(void)
{
  printf("octets %d / %d ; pointeurs %d / %d ; realloc %d\n",
          v_dbg_alloc_bytes, v_dbg_free_bytes, v_dbg_alloc_count, v_dbg_free_count, v_dbg_realloc_count);
}

//Reset des compteurs d'allocation
void f_dbg_end_alloc_count(void)
{
  if (v_dbg_mem_on == 0)
    printf("WARNING! Pas de comptage mmoire en cours.\n");
  else
  {
    printf("Octets allous/librs: %d / %d (max=%d)\n",
           v_dbg_alloc_bytes, v_dbg_free_bytes, v_dbg_bytes_max);
    printf("Pointeurs allous/rallous/librs: %d / %d / %d (max=%d)\n",
           v_dbg_alloc_count, v_dbg_realloc_count, v_dbg_free_count, v_dbg_count_max);
  }
  v_dbg_mem_on = 0;
}

//Pour malloc --> appeler la macro m_dbg_malloc
void *f_dbg_malloc(int size)
{
  t_dbg_alloc_block *pb;
  //Dans l'allocation, prvoir la taille des infos de debug
  pb = (t_dbg_alloc_block*) malloc(M_DBG_ALLOC_HEADER_SIZE + size);
  if (pb == NULL)
  {
    printf("f_dbg_malloc(%d): chec de l'allocation\n", size);
    return NULL;
  }

  //Allocation russie:
  pb->prefix = M_DBG_ALLOC_HEADER_PREFIX;
  pb->size = size;

  //Mettre les compteurs  jour
  if (v_dbg_mem_on == 1)
  {
    v_dbg_alloc_count++;
    v_dbg_alloc_bytes+=size;
    if ( (v_dbg_alloc_bytes - v_dbg_free_bytes) > v_dbg_bytes_max )
      v_dbg_bytes_max = v_dbg_alloc_bytes - v_dbg_free_bytes;
    if ( (v_dbg_alloc_count - v_dbg_free_count) > v_dbg_count_max )
      v_dbg_count_max = v_dbg_alloc_count - v_dbg_free_count;

    if (v_dbg_alloc_bytes < 0)
      printf("DEBUG:f_dbg_malloc(%06d): v_dbg_alloc_bytes=%d\n", size, v_dbg_alloc_bytes);
  }

  //Renvoyer pointeur vers 'contents'
#ifdef DEBUG_MEM_VERBOSE
  printf("f_dbg_malloc(%06d) --> 0x%08x\n", size, (int) &pb->contents);
#endif
  return &pb->contents;
}

//Pour 'realloc' --> appeler la macro m_dbg_realloc
void *f_dbg_realloc(void *p, int new_size)
{
  char *q;
  int old_size;
  t_dbg_alloc_block *pb, *new_pb;

  IFDBG { printf("f_dbg_realloc()"); }

  //Si p=NULL on fait comme avec m_alloc
  if (p == NULL)
    return m_dbg_malloc(new_size);

  //Sinon ... trouver le dbut du bloc rellement allou
  q = (char*) p;
  q -= M_DBG_ALLOC_HEADER_SIZE;
  pb = (t_dbg_alloc_block*) q;

  //Vrifier si on retrouve la taille et le motif attendus.
  if (pb->prefix != M_DBG_ALLOC_HEADER_PREFIX)
  {
    printf("m_dbg_realloc: prefix manquant en 0x%08x", (int) pb);
    if (pb->prefix == M_DBG_FREE_HEADER_PREFIX)
      printf("(pointeur 0x%08x dj libr?)\n", (int) p);
    else
     printf("(mauvais pointeur 0x%08x?)\n", (int) p);
  }

  //Rallocation  la faon du C
  old_size = pb->size;
  if (new_size == 0)
  {
    //Nouvelle taille nulle: vider le pointeur
    f_dbg_free(p);
    return NULL;
  }
  else
  {
    new_pb = (t_dbg_alloc_block*) realloc(pb, M_DBG_ALLOC_HEADER_SIZE+new_size);
    if (new_pb == NULL)
    {
      //Erreur
      printf("m_dbg_realloc: 0x%08x erreur interne sur 0x%08x.\n", (int) p, (int) pb);
      return NULL;
    }
    else {
      //Rallocation russie: crire la nouvelle taille
      //(note: ncessaire mme si new_pb=pb)
      new_pb->size = new_size;
      //Rallocation russie: actualiser les compteurs
      if (v_dbg_mem_on != 0)
      {
        v_dbg_realloc_count++;
        v_dbg_alloc_bytes += new_size - old_size;
        if ( (v_dbg_alloc_bytes - v_dbg_free_bytes) > v_dbg_bytes_max )
          v_dbg_bytes_max = v_dbg_alloc_bytes - v_dbg_free_bytes;
        if (v_dbg_alloc_bytes < 0)
          printf("DEBUG:f_dbg_realloc(%d): v_dbg_alloc_bytes=%d\n", new_size, v_dbg_alloc_bytes);
      }
      //...et renvoyer la zone mmoire rserve
      return &new_pb->contents;
    }
  }
}

//Pour free --> appeler la macro m_dbg_free
void f_dbg_free(void *p)
{
  char *q; //Intermdiaire
  t_dbg_alloc_block *pb;
  int size;

  //Dallocation de NULL: ne  rien faire
  if (p == NULL)
    return;

  //Retrouver le bloc de debug
  q = (char*) p;
  q -= M_DBG_ALLOC_HEADER_SIZE;
  pb = (t_dbg_alloc_block *) q;

  //Contrle des prfixes
  if (pb->prefix != M_DBG_ALLOC_HEADER_PREFIX)
  {
    printf("f_dbg_free: 0x%08x --> 0x%08x mauvais prfixe", (int) p, (int) pb);
    if (pb->prefix == M_DBG_FREE_HEADER_PREFIX)
      printf("(double libration probable)\n");
    else
      printf("(crasement ou mauvais pointeur)\n");
  }
#ifdef DEBUG_MEM_VERBOSE
  printf("f_dbg_free(0x%08x)\n", p);
#endif

  //Libration du bloc
  size = pb->size;
  pb->prefix = M_DBG_FREE_HEADER_PREFIX;
  free(pb);

  //Mise  jour des compteurs
  if (v_dbg_mem_on == 1)
  {
    v_dbg_free_bytes += size;
    v_dbg_free_count++;

    //Erreur potentielle
    if (v_dbg_free_count > v_dbg_alloc_count)
      printf("f_dbg_free: DEBUG %d librations > %d allocations\n", v_dbg_free_count, v_dbg_alloc_count);
  }

  return;
}

// ----- Rservation de tableaux dynamiques multi-dimensions -----

//Arguments:
//  dim : nombre de dimensions
//  es  : 'Element Size': taille d'un lment, en octets
//  ... :  'dim' entiers donnant les dimensions du tableau
//Retour:
//  pointeur vers le tabgleau si l'allocation a russi.
//  NULL sinon.
void *f_dbg_mult_alloc(int dim, int es, ...)
{
  va_list ap;

  va_start(ap,es);

  return f_dbg_mult_alloc_aux(dim, es, ap);
}

//Implmentation rcursive de f_dbg_mult_alloc, avec 'ap' au lieu de ...
static void *f_dbg_mult_alloc_aux(int dim, int es, va_list ap)
{
  int i, d;
  void *p;
  void **a;

  //Une seule dimension: facile!
  if (dim == 1) {
    d = va_arg(ap,int);
    p = (void*) m_dbg_malloc(es * d);
    return p;
  //Plusieurs dimensions: rserver chaque partie de tableau.
  } else {
    d = va_arg(ap,int);
    //A priori c'est un tableau de pointeurs
    a = (void**) m_dbg_malloc(sizeof(void*) * d);
    if (a == NULL)
      return a;
    //Allouer chaque pointeur
    for(i=0; i<d; i++) {
      p = (void*) f_dbg_mult_alloc_aux(dim-1, es, ap);
      if (p == NULL)
        goto clean_and_leave;
      a[i] = p;
    }
    //Si on a pu allouer chaque lment: succs
    return (void*) a;
  }

  clean_and_leave:
    i--;
    while (i >= 0) {
      f_dbg_mult_free_aux(a[i], dim-1, ap);
    }
    return NULL;
}

// ------ Libration de tableaux dynamiques multi-dimensions -----

//Libre un tableau multi-dimensions
//Arguments:
//  p   : le tableau en question
//  dim : nombre des dimensions du tableau.
//  ... : valeurs des dimensions du tableau.
void f_dbg_mult_free(void *p, int dim, ...)
{
  va_list ap;
  va_start(ap, dim);

  f_dbg_mult_free_aux(p, dim, ap);
}

//Fonction auxiliaire du prcdent.
//Idem mais avec 'ap' au lieu de ...
static void f_dbg_mult_free_aux(void *p, int dim, va_list ap)
{
  int i, d;
  void **q;

  //Une seule dimension
  if (dim == 1) {
    m_dbg_free(p);
  //Plus de dimensions
  } else {
    d = va_arg(ap, int); //Dsormais 'ap' pointe la dimension qui suit
    q = (void**) p;
    for(i=0; i<d; i++) {
       f_dbg_mult_free_aux(q[i], dim-1, ap);
    }
    m_dbg_free(p);
  }
}

#endif //#ifdef DEBUG_MEM

// ----- Copie de tableaux dynamiques multi-dimensions -----

//Note: on suppose que les rservations pour 'tgt' sont DEJA faites.
//Arguments:
//  dim : nombre de dimensions (>=1)
//  tgt : zone mmoire cible
//  src : zone mmoire source
//  es  : 'Element Size': taille d'un lment, en octets
//  ... :  'dim' entiers donnant les dimensions du tableau
void f_dbg_mult_memcpy(void *tgt, void *src, int dim, int es, ...)
{
  va_list ap;

  va_start(ap,es);

  f_dbg_mult_memcpy_aux(tgt, src, dim, es, ap);
}

//Idem, avec 'va_list' dans les arguments.
static void f_dbg_mult_memcpy_aux(void *tgt, void *src, int dim, int es, va_list ap)
{
  int d;
  d = va_arg(ap,int); //Lit la dimension et avance 'ap' d'un cran.

  //Une seule dimension
  if (dim == 1)
  {
    memcpy(tgt, src, es*d);
  }
  //Plusieurs dimensions
  else
  {
    int i;
    void **tgt2, **src2;

    tgt2 = (void**) tgt;
    src2 = (void**) src;

    for(i=0; i<d; i++)
    {
      f_dbg_mult_memcpy_aux(tgt2[i], src2[i], dim-1, es, ap);
    }
  }
}

// ----- Comparaison de tableaux dynamiques de mmes dimensions -----

//Compare deux tableaux multi-dimensionnels.
//Arguments:
//  dim : nombre de dimensions (>=1)
//  a0  : Tableau
//  a1  : Tableau  comparer (suppos de mme dimension)
//  es  : 'Element Size': taille d'un lment, en octets
//  ... :  'dim' entiers donnant les dimensions du tableau
int f_dbg_mult_memcmp(void *a0, void *a1, int dim, int es, ...)
{
  va_list ap;
  va_start(ap,es);
  return f_dbg_mult_memcmp_aux(a0, a1, dim, es, ap);
}

//Idem avec 'ap' au lieu de '...'
static int f_dbg_mult_memcmp_aux(void *a0, void *a1, int dim, int es, va_list ap)
{
  int d;
  d = va_arg(ap,int); //Lit la dimension et avance 'ap' d'un cran.

  //Une seule dimension
  if (dim == 1)
  {
    return memcmp(a0, a1, es*d);
  }
  //Plusieurs dimensions
  else
  {
    int i,r;
    void **b0, **b1;

    b0 = (void**) a0;
    b1 = (void**) a1;

    for(i=0; i<d; i++)
    {
      r = f_dbg_mult_memcmp_aux(b0[i], b1[i], dim-1, es, ap);
      if (r != 0)
        return r;
    }
    return 0;
  }
}

// ===== Fin de fichier =====
