// =================================================
// Dclarations gnrales pour faciliter le dbogage
//
//   m_dbg_malloc : Comme 'malloc' sauf si DEBUG_MEM est dfini
//   m_dbg_free   : Comme 'free'   sauf si DEBUG_MEM est dfini
//   IFDBG { }    : instructions pour le dbogage
//   void f_dbg_start_alloc_count(void); : compte les appels  m_dbg_malloc
//   void f_dbg_end_alloc_count(void);   : affiche le nombre de pointeurs allous/librs.
//
// Options:
//   DEBUG_ON     : passer dans les blocs IFBG si DEBUG_ON vaut 0.
//   DEBUG_MEM    : m_dbg_malloc et m_dbg_free appelleront les versions "espionnes" de malloc et free.
//   DEBUG_MEM_VERBOSE : chaque appel  m_dbg_malloc ou m_dbg_free sera affich.
//
// =================================================

#ifndef _DEBUG_H_
#define _DEBUG_H_

//Dans le code  dboguer on peut procder ainsi:
//...
//  IFDBG { instructions_de_debug }
//...
//puis dfinir DEBUG_ON  1 avant le code  dboguer et  0 aprs.
//
#define DEBUG_ON 0
#define IFDBG if ( DEBUG_ON )

//Choisit si on veut dboguer les allocations/librations de mmoire
#define DEBUG_MEM
//Fait un "printf" sur chaque allocation/libration de pointeur.
//#define DEBUG_MEM_VERBOSE

#ifdef DEBUG_MEM
//Macros pour redirection
#define m_dbg_malloc(_a_)         f_dbg_malloc(_a_)
#define m_dbg_realloc(_a_,_b_)    f_dbg_realloc(_a_,_b_)
#define m_dbg_free(_a_)           f_dbg_free(_a_)
#define m_dbg_start_alloc_count() f_dbg_start_alloc_count()
#define m_dbg_end_alloc_count()   f_dbg_end_alloc_count()
#define m_dbg_print_counters()    f_dbg_print_counters()
//Fonctions rediriges alloc/free
void *f_dbg_malloc(int size);
void *f_dbg_realloc(void *p, int new_size);
void f_dbg_free(void *p);
void f_dbg_start_alloc_count(void);
void f_dbg_end_alloc_count(void);
void f_dbg_print_counters(void);
#else
//Pas de dbogage alloc/free
#define m_dbg_malloc(_a_)       malloc(_a_)
#define m_dbg_realloc(_a_,_b_)  realloc(_a_,_b_)
#define m_dbg_free(_a_)         free(_a_)
#define m_dbg_start_alloc_count()
#define m_dbg_end_alloc_count()
#define m_dbg_print_counters()
#endif

// ----- Services de dbogage mmoire permanents -----

// ----- Services d'allocation multi-dimensionnelle -----

//Fonctions d'allocation/libration/copie/comparaison gnriques
void *f_dbg_mult_alloc(int dim, int es, ...); //Tableau  'dim' dimensions, lments de taille 'es', <dimensions>
void f_dbg_mult_free(void *p, int dim, ...);  //Librer le tableau.
void f_dbg_mult_memcpy(void *tgt, void *src, int dim, int es, ...); //Copier les tableaux
int f_dbg_mult_memcmp(void *a0, void *a1, int dim, int es, ...);    //Comparer les tableaux

//Macros d'allocation  2,3,4 dimensions
#define m_dbg_malloc2(_t_,_d1_,_d2_)           ((_t_**)   f_dbg_mult_alloc(2,sizeof(_t_),_d1_,_d2_))
#define m_dbg_malloc3(_t_,_d1_,_d2_,_d3_)      ((_t_***)  f_dbg_mult_alloc(3,sizeof(_t_),_d1_,_d2_,_d3_))
#define m_dbg_malloc4(_t_,_d1_,_d2_,_d3_,_d4_) ((_t_****) f_dbg_mult_alloc(4,sizeof(_t_),_d1_,_d2_,_d3_,_d4_))

//Macros de libration  2,3,4 dimensions
#define m_dbg_mfree2(_p_,_d1_,_d2_)           f_dbg_mult_free(_p_,2,_d1_,_d2_)
#define m_dbg_mfree3(_p_,_d1_,_d2_,_d3_)      f_dbg_mult_free(_p_,3,_d1_,_d2_,_d3_)
#define m_dbg_mfree4(_p_,_d1_,_d2_,_d3_,_d4_) f_dbg_mult_free(_p_,4,_d1_,_d2_,_d3_,_d4_)

#endif
// ===== Fin de fichier =====


