/***************************************************************************/
/* IHM POUR SUDOKU                                                         */
/***************************************************************************/

// ===== Includes =====

// ----- Librairies standard -----
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>

// ----- Sources du projet -------

#include "debug.h"

#include "sudoku.h"
#include "algorithme.h"

#include "sudoku_funcs.h"
#include "algorithme_funcs.h"

#include "ihm.h"

#include "valid.h"

//===========================================================================
// Types et structures
//===========================================================================

// ==========================================================================
// Variables globales
// ==========================================================================

int ihm_Boardsize; // ct d'une grille
int ihm_Elemsize;  // ct d'une rgion carre de la grille
int ihm_Maxcoul;   // valeurs autorises: 1..ihm_Maxcoul
bool_t ihm_b_log = FALSE;  // TRUE si fichier de log
char *ihm_s_log  = "su_log.txt"; // Fichier de log par dfaut
FILE *ihm_v_file;

bool_t IHM_v_opt_balai;    //Utiliser le balayage
bool_t IHM_v_opt_unique;   //Rechercher les cases  coup unique
bool_t IHM_v_opt_nettoyer; //Eliminer des coups par ensemble
bool_t IHM_v_opt_impasse;  //Reconnatre une impasse quand il y a 0 coups sur une case.
bool_t IHM_v_opt_expliquer;//Expliquer la dmarche de recherche

// ==========================================================================
// Prototypes
// ==========================================================================

static void IHM_f_clear(JEU_position_t *p_position);

//===========================================================================
// Routines d'interface et de test
//===========================================================================

//Active le fichier de log
void ihm_log_start(void)
{
  if (ihm_b_log == FALSE)
  {
    printf("LOG: dmarrage: %s\n", ihm_s_log);
    ihm_v_file = fopen(ihm_s_log, "w");
    ihm_b_log = TRUE;
  }
}

//Dsactive le fichier de log
void ihm_log_end(void)
{
  if (ihm_b_log == TRUE)
  {
    printf("LOG: arrt\n");
    fclose(ihm_v_file);
    ihm_b_log = FALSE;
  }
}

//Comme "printf", mais avec fichier de log (ou non) suivant options.
void ihm_out(const char *format, ... )
{
  va_list ap;
  va_start(ap,format);
  vprintf(format, ap); //Sortie standard
  if (ihm_b_log == TRUE)
  {
    vfprintf(ihm_v_file, format,ap);//Fichier de log
  }
  va_end(ap);
}

void ihm_i2xy(int i, int *px, int *py)
{
  *px = i % ihm_Boardsize;
  *py = i / ihm_Boardsize;
}

void ihm_getcarre(int *px, int *py)
{
  *px = (*px) - ((*px) % ihm_Elemsize);
  *py = (*py) - ((*py) % ihm_Elemsize);
}

int ihm_xy2i(int x, int y)
{
  return ((ihm_Boardsize*y)+x);
}


/* Ligne de sparation horizontale pour grille de sudoku */
/* bs: ct de la grille */
/* es: ct d'un carr (NB: es*es=bs) */
void ihm_drawline(int bs, int es)
{
  int x;
  for(x=0; x<bs; x++)
    {
      if (x % es == 0)
        ihm_out("+");
      ihm_out("--");
    }
  ihm_out("+\n");
}

/* Affiche une grille de sudoku */
/* Colonnes numrotes  partir de A */
/* Lignes numrotes  partir de 0 */
void ihm_affiche(JEU_position_t *p_position)
{
  int y, x;
  int v;
  int bs, es; //BoardSize, ElemSize

  bs = p_position->Boardsize;
  es = p_position->Elemsize;

  //Nom des colonnes
  ihm_out("  ");
  for(x=0; x<bs; x++)
    {
      if (x % es == 0)
        ihm_out(" ");
      ihm_out("%c ", (char)((int) 'A' + x) );
    }
  ihm_out("\n");
  //Contenu du tableau
  for(y=0; y<bs; y++)
    {
      if (y % es == 0)         //Ligne en bas des carrs
        {
          ihm_out(" ");
          ihm_drawline(bs,es);
        }
      ihm_out("%d", y);         //Numro de ligne
      for (x=0; x<bs; x++)     //Contenu de la ligne
        {
          if (x % es == 0)
            ihm_out("|");
          v = p_position->cases[x][y];
          if (v != 0)
            ihm_out("%2d", v);
          else
            ihm_out("  ");
        }
      ihm_out("|\n");
    }
  //Bord infrieur
  ihm_out(" ");
  ihm_drawline(bs, es);
}

/* Explique comment spcifier les options */
void ihm_fp_explique_options(void)
{
	printf("---------------------------------------------\n");
  printf("Choix des options:\n");
  printf("  b:balayage\n");
  printf("  u:coup unique\n");
  printf("  n:nettoyage\n");
  printf("  i:impasses\n");
  printf("  e:explications\n");
  printf("puis '+' ou '-' pour activer, ou dsactiver\n");
  printf("  'h': aide\n");
  printf("  'v': voir les options\n");
  printf("  'q': valider\n");
  printf("---------------------------------------------\n");
}

/* Affiche les options courantes */
void ihm_fp_affiche_options(void)
{
  printf("OPTIONS: balayage=%d, coup_unique=%d, nettoyage=%d, impasse=%d, explications=%d\n",
          IHM_v_opt_balai, IHM_v_opt_unique, IHM_v_opt_nettoyer, IHM_v_opt_impasse, IHM_v_opt_expliquer);
}

/* Demande  l'utilisateur d'activer ou de dsactiver des options */
void ihm_fp_saisir_options(void)
{
  bool_t v;
  int L;
  char s[3];

  ihm_fp_explique_options();

  while(TRUE)
  {
    fgets(s, 3, stdin);

    //Activer ou dsactiver une option?
    L = strlen(s);
    if (L >= 2)
    {
      switch(s[1])
      {
        case '+':
          v = TRUE;
          break;
        case '-':
          v = FALSE;
          break;
      }
    }

    //De quelle option s'agit-t-il?
    if (L >= 1)
    {
      switch(s[0])
      {
        case 'b':
          IHM_v_opt_balai = v;
          break;
        case 'e':
          IHM_v_opt_expliquer = v;
          break;
        case 'u':
          IHM_v_opt_unique = v;
          break;
        case 'n':
          IHM_v_opt_nettoyer = v;
          break;
        case 'i':
          IHM_v_opt_impasse = v;
          break;
        case 'q':
          ihm_fp_affiche_options();
          return;
        case 'h':
          ihm_fp_explique_options();
          break;
        case 'v':
          ihm_fp_affiche_options();
          break;
        case '\d': //Fin de ligne Windows
        case '\n': //Fin de ligne Unix
          break;
        default:
          printf("ERREUR: %c n'est pas une commande pour ce menu\n", s[0]);
          break;
      }
    }
    fflush(stdin);
  }
}

/* Remplit la ligne n de 'p_position' avec les caractres de 's' */
void IHM_f_priv_fill_ligne(char *s, int n, JEU_position_t *p_position)
{
  int bs;
  int i,x, y, L;
  L = strlen(s);

  bs = p_position->Boardsize;

  x = 0;
  y = n;
  for(i=0; i<L; i++)
  {
    char c;
    c = s[i];
    switch(c)
    {
      case '0':
      case ' ':
        p_position->cases[x][y] = 0;
        x +=1;
        break;
      case '1':
      case '2':
      case '3':
      case '4':
      case '5':
      case '6':
      case '7':
      case '8':
      case '9':
        p_position->cases[x][y] = (char) (((int) c) - ((int) '0'));
        x +=1;
        break;
    }
    //Ligne trop longue?
    if (x >= bs)
      break;
  }
  //Ligne trop courte?
  while(x<bs)
  {
    p_position->cases[x][y] = 0;
    x += 1;
  }
}

/* Interprte la ligne de commande fournie par l'utilisateur */
/*   c<n>: corriger la saisie de la ligne <n>
/*   v   : voir
/*   <n>...   : remplissage de la ligne courante
/*   <spc>... : remplir la ligne en commenant par une case vide
/*   q   : quitter
/*   rien: Ligne vide
/*Arguments:
/* p_position: les cases  remplir
/* n         : index de la ligne courante
/* s         : la commande de l'utilisateur
/*
/* Renvoie la taille de la grille si l'utilisateur a valid la grille,
/* sinon (-1).
*/
int IHM_f_interprete(char *s, int n, JEU_position_t *p_position)
{
  int bs;
  int n1;
  char c;
  int L;

  bs = p_position->Boardsize;
  L = strlen(s);

  //Ligne vide ne contenant que "\n"
  if (L == 1)
  {
    IHM_f_priv_fill_ligne(s, n, p_position);
    n += 1;
  }
  //Ligne non-vide ou commande utilisateur.
  else
  {
    c = s[0];
    switch(c)
    {
      case 'c': //Corriger une ligne
        if (L >= 1)
        {
          n1 = ((int) s[1]) - ((int) '0');
          if ((n1<bs) && (n1>=0))
            n=n1;
        }
        break;
      case 'v': //Voir o on en est
        ihm_affiche(p_position);
        break;
      case ' ':
      case '0': //Remplissage: <espace>0123456789
      case '1':
      case '2':
      case '3':
      case '4':
      case '5':
      case '6':
      case '7':
      case '8':
      case '9':
        IHM_f_priv_fill_ligne(s, n, p_position);
        n += 1;
        break;
      case 'd': //Dmonstration: affiche une grille prdfinie
        ihm_saisir2(p_position);
        ihm_affiche(p_position);
        break;
      case 'h': //Aide
        printf("------------------------------------------\n");
        printf("Aide:\n");
        printf("  Saisissez le contenu de chaque ligne en terminant par 'ENTREE'\n");
        printf("  Pour une case vide, tapez un espace ou 0\n");
        printf("  Tapez 'q' pour quitter, 'r' pour rsoudre la grille, 'v' pour l'afficher,\n");
        printf("  'o' pour les options de recherche\n");
        printf("  Vous pouvez resaisir la ligne <n> en tapant c<n>\n");
        printf("  Pour prendre une grille pr-remplie, tapez d\n");
        printf("  Enfin, tapez 'h' pour rafficher ce message.\n");
        printf("  Activation/Dactivation du log: L\n");
        printf("------------------------------------------\n");
        break;
      case 'L': // Activation/Dactivation du log
        if (ihm_b_log == FALSE)
          ihm_log_start();
        else
          ihm_log_end();
        break;
      case 'o': //Saisie des options
        ihm_fp_saisir_options();
        break;
      case 'q': //Quitter
        return (-1);
        break;
      case 'r': //Rechercher la solution
        ihm_affiche(p_position);
        ALG_f_resoudre(p_position);
        break;
      default:  //Ignorer tout le reste.
        break;
    }
  }
  //On ne quitte pas ...
  if (n>=bs)
    n = bs-1;
  return n;
}

/* Demande de saisir une grille ... de sudoku */
/* Renvoie TRUE si la saisie est valide, FALSE pour abandon*/
bool_t IHM_f_saisir(JEU_position_t *p_position)
{
  int x,y;
  int bs, es; //Abbrviations pour "Boardsize", "Elemsize"
  char s[80];

  //Options de recherche par dfaut
  IHM_v_opt_balai    = TRUE;  //Utiliser le balayage
  IHM_v_opt_unique   = TRUE;  //Rechercher les cases  coup unique
  IHM_v_opt_nettoyer = TRUE;  //Eliminer des coups par ensemble
  IHM_v_opt_impasse  = TRUE;  //Reconnatre une impasse quand il y a 0 coups sur une case.
  IHM_v_opt_expliquer= TRUE;  //Expliquer la dmarche de recherche

  bs = 9; //Ct du plateau
  es = 3; //Ct d'un carr
  p_position->Boardsize = bs;
  p_position->Elemsize  = es;

  //Rservation mmoire
  p_position->cases = (unsigned char**) m_dbg_malloc(bs*sizeof(unsigned char*));
  for(x=0; x<bs; x++)
    p_position->cases[x] = (unsigned char*) m_dbg_malloc(bs*sizeof(unsigned char));

  //Interface utilisateur
  printf("Saisir la grille. Taper 'h' pour de l'aide:\n");
  y = 0;
  while((y>=0) && (y<bs))
  {
    fflush(stdin);
    //scanf("%s", s);
    fgets(s, 80, stdin);
    y=IHM_f_interprete(s, y, p_position);
  }

  //Libration mmoire
  IHM_f_clear(p_position);

  return TRUE;
}

static void IHM_f_clear(JEU_position_t *p_position)
{
  int x;
  int bs; //Abbrviations pour "Boardsize"

  bs = p_position->Boardsize;

  for(x=0; x<bs; x++)
     m_dbg_free(p_position->cases[x]);
  m_dbg_free(p_position->cases);
}

// Affiche piste explore actuellement
void ihm_PrintExplo(ALG_plan_explore_t *pexplo, int idx_explo)
{
  int ic;
  ALG_list_coups_t *pcL;

  pcL = &pexplo->coups_cases[idx_explo];
  for(ic=0; ic<pcL->nb_coups; ic++)
  {
    //ihm_out("%d", lc->coups[i]);
  }
  ihm_out(")\n");
}

// Grille prdfinie pour debug
void ihm_saisir2(JEU_position_t *p_position)
{
// ----- grille "Diabolique" initiale
  unsigned char tab0[9][9] = {{0,0,0, 0,0,0, 2,5,0},
                              {0,0,4, 0,2,9, 0,3,0},
                              {0,3,0, 0,0,4, 0,0,0},

                              {0,0,0, 2,4,0, 0,0,7},
                              {0,2,0, 7,0,0, 8,0,0},
                              {0,9,3, 0,0,0, 0,0,0},

                              {8,0,0, 0,5,0, 0,0,0},
                              {5,6,0, 0,0,0, 0,0,2},
                              {0,0,0, 1,0,0, 0,6,0}};

// ----- grille "Diabolique" avec tatonnement rsolu B4=8 -----


//unsigned char tab0[9][9] = {{0,0,0, 3,0,0, 2,5,0},
//                            {0,0,4, 0,2,9, 0,3,0},
//                            {0,3,0, 0,0,4, 0,0,0},
//
//                            {0,8,0, 2,4,0, 0,0,7},
//                            {0,2,0, 7,0,0, 8,0,0},
//                            {0,9,3, 0,0,0, 0,0,0},
//
//                            {8,0,0, 0,5,0, 0,0,0},
//                            {5,6,0, 0,0,0, 0,0,2},
//                            {0,0,0, 1,0,0, 0,6,0}};

// ----- grille "Diabolique" rsolue -----

//   solution = {{9,1,8, 3,6,7, 2,5,4},
//               {6,5,4, 8,2,9, 7,3,1},
//               {2,3,7, 5,1,4, 6,8,9},
//
//               {1,8,6, 2,4,5, 3,9,7},
//               {4,2,5, 7,9,3, 8,1,6},
//               {7,9,3, 6,8,1, 4,2,5},
//
//               {8,7,2, 9,5,6, 1,4,3},
//               {5,6,1, 4,3,8, 9,7,2},
//               {3,4,9, 1,7,2, 5,6,8}};

// ----- Grille sur le site du "Monde"

unsigned char tab1[9][9] = {{8,0,0, 0,7,9, 0,0,0},
                            {0,7,3, 0,0,0, 0,0,2},
                            {0,0,0, 6,0,0, 0,4,3},

                            {0,0,0, 0,0,0, 0,0,0},
                            {1,0,0, 2,0,0, 3,0,0},
                            {0,9,0, 0,0,5, 0,0,4},

                            {3,1,0, 0,0,0, 0,0,0},
                            {0,0,6, 0,0,0, 5,0,8},
                            {0,0,0, 0,5,0, 4,0,0}};

// +------+------+------+
// | 8    |   7 9|      |
// |   7 3|      |     2|
// |      | 6    |   4 3|
// +------+------+------+
// |      |      |      |
// | 1    | 2    | 3    |
// |   9  |     5|     4|
// +------+------+------+
// | 3 1  |      |      |
// |     6|      | 5   8|
// |      |   5  | 4    |
// +------+------+------+
//   0: Balayage Carr:(3,0)=3; (3,1)=5; (6,2)=7; (1,3)=3; (4,5)=3; (2,6)=5;
//   6: Nettoyage sur un carr:
//      L'ensemble (6,0)(7,0)(8,0) couvre 156
//   7: Balayage Ligne:(0,1)=6;  ==> ici un balayage "carr" devrait suffire.
//
// ----- Grille site web des ings Tunisiens -----
// +------+------+------+
// | 1 4  |      |     6|
// |     9| 3   6|      |
// |      |   7 1|      |
// +------+------+------+
// |   7 1|     5|      |
// |      | 8    | 7    |
// |     3|   1  | 9 5  |
// +------+------+------+
// | 7    | 9 5  |   3 8|
// |   5  |      | 1 4  |
// | 6    |      |     9|
// +------+------+------+
// 2 solutions trouves: qui a tort?
//   0 : Balayage Carr:(2,0)=7; (0,7)=9; (1,4)=9; (4,3)=9; (5,0)=9; (7,2)=9; (1,8)=3; (0,2)=3; (6,0)=3; (1,6)=1; (3,8)=1; (6,8)=5; (6,6)=6;
//   13: Balayage Ligne:(3,0)=5; (8,3)=3;
//   ...
// +------+------+------+
// | 1 4 7| 5 2 9| 3 8 6|
// | 5 2 9| 3 8 6| 4 7 1|
// | 3 8 6| 4 7 1| 2 9 5|
// +------+------+------+
// | 4 7 1| 2 9 5| 8 6 3|
// | 2 9 5| 8 6 3| 7 1 4|
// | 8 6 3| 7 1 4| 9 5 2|
// +------+------+------+
// | 7 1 4| 9 5 2| 6 3 8|
// | 9 5 2| 6 3 8| 1 4 7|
// | 6 3 8| 1 4 7| 5 2 9|
// +------+------+------+
// +------+------+------+
// | 1 4 7| 5 2 9| 3 8 6|
// | 5 2 9| 3 8 6| 4 1 7|
// | 3 8 6| 4 7 1| 2 9 5|
// +------+------+------+
// | 4 7 1| 6 9 5| 8 2 3|
// | 2 9 5| 8 3 4| 7 6 1|
// | 8 6 3| 2 1 7| 9 5 4|
// +------+------+------+
// | 7 1 4| 9 5 2| 6 3 8|
// | 9 5 8| 7 6 3| 1 4 2|
// | 6 3 2| 1 4 8| 5 7 9|
// +------+------+------+

  int x,y;
  int bs; //Abrviations Boardsize, Elemsize

  bs = 9;

  for(x=0; x<bs; x++)
    for(y=0; y<bs; y++)
      p_position->cases[x][y] = tab1[y][x];
}

//===========================================================================
// Programme de test
//===========================================================================

int main(void)
{

  JEU_position_t position;

  IHM_f_saisir(&position);
  ////Validation des routines
  //f_valid_multialloc();

  return 0;
}
