/***************************************************************************/
/* SUDOKU                                                                  */
/***************************************************************************/

// ===== Includes =====

// ----- Librairies standard -----
#include <stdlib.h>
#include <stdio.h>

// ----- Sources du projet -------

#include "debug.h"

#include "sudoku.h"
#include "algorithme.h"

#include "sudoku_funcs.h"
#include "algorithme_funcs.h"

#include "ihm.h"

#include "valid.h"

//===========================================================================
// Types et structures
//===========================================================================

// ------------------ Types propres au jeu ----------------------------------
// Nommage:
//  _t    : suffixe pour tous les types
//  JEU_  : prfixe soulignant l'aspect spcifique au jeu
//  _priv_: type 'priv' non-utilis par des routines externes au jeu.
//  i : index
//  a : tableau
//  n : valeur
//  e : etat

//Bilans sur un ensemble de cases
//  i  : index d'une case
//  2  : 'to' correspondance
//  nb : mesure de quantit
//  n  : valeur joue ou  jouer
//  e  : etat
typedef enum { JEU_SET_LIGNE_E, JEU_SET_COLONNE_E, JEU_SET_CARRE_E } JEU_set_type_t;

typedef struct
{
  JEU_set_type_t t;
  int *a_i2x; //Index case --> coordonnes
  int *a_i2y;
  //int *a_n2nb; // a_n2nb[n]   : Nombre de cases LIBRES o on peut jouer le coup 'n'
  int *a_n2e;  // a_n2e[n]    : Nombre de cases occupes par N
}
JEU_priv_cases_set_t;

//Etat  prendre en considration pour avancer/reculer
typedef struct
{
  JEU_position_t plateau;       // Les infos affichables

  // Bilans  actualiser pour chaque coup
  //'Boardsize' ensembles * 3 (ligne, colonne, carr)
  JEU_priv_cases_set_t *a_sets; //Indique quelles cases sont occupes, et par quoi.

  char ***a_xyn2e; // a_xy2n[x][y][n] := 0..3 (occurrences de n dans ligne, colonne, carr de la case x,y, occupe ou non)
  char **a_xy2nb;  // a_xy2nb[x][y]   := Nombre d'indexs nuls dans .a_xyn2e[]

  int Boardsize; // Recopi depuis "plateau"
  int Maxcoul;
  int Elemsize;
}
JEU_priv_etat_t;

// ====== Variables globales prives =====

static JEU_priv_etat_t JEU_v_priv_etat;

JEU_explain_t JEU_v_priv_last_cause; //Pour l'affichage des explications
int           JEU_v_priv_last_idx;

// ====== Prototypes locaux =====

// ----- pour le dbogage -----

static void JEU_f_priv_print_subset(JEU_priv_cases_set_t *ps, char *a_subset, char *a_vals);
static void JEU_f_priv_list_set(JEU_priv_cases_set_t *ps);
static void JEU_f_priv_print_set(JEU_priv_cases_set_t *ps);

// ----- pour le jeu -----

static JEU_priv_cases_set_t* JEU_f_priv_get_set_col(int x);     //Conversion coordonnes-->index des ensembles
static JEU_priv_cases_set_t* JEU_f_priv_get_set_ligne(int y);
static JEU_priv_cases_set_t* JEU_f_priv_get_set_carre(int x, int y);

static int JEU_f_priv_mem_set_col(int x); //Rservations de mmoire et coordonnes pour un ensemble
static int JEU_f_priv_mem_set_ligne(int y);
static int JEU_f_priv_mem_set_carre(int x, int y);

static void JEU_f_priv_mem_set_free(JEU_priv_cases_set_t *ps); //Rend la mmoire rserve pour un ensemble

static int JEU_f_priv_get_set_n2nb(JEU_priv_cases_set_t *ps, int n); //Compte les cases possibles pour 'n'

static bool_t JEU_f_priv_init_part_set(JEU_priv_cases_set_t *ps);  //Met  jour a_xyn2e[x,y,v] pour toutes les cases de *ps.

//====== Actions lies au jeu =====

// ---- Jeu: dbogage -----

static void JEU_f_priv_print_subset(JEU_priv_cases_set_t *ps, char *a_subset, char *a_vals)
{
  int bs;
  int i, x, y, v;

  bs = JEU_v_priv_etat.Boardsize;

  ihm_out("L'ensemble ");
  for(i=0; i<bs; i++)
    {
      if (a_subset[i] != 0)
        {
          x = ps->a_i2x[i];
          y = ps->a_i2y[i];
          ihm_out("%c%d,", (char) (((int) 'A') + x),y);
        }
    }
  ihm_out(" couvre ");
  for(v=1; v<=bs; v++)
    {
      if (a_vals[v] != 0)
        ihm_out("%d", v);
    }
  ihm_out("\n");
}

//Affiche les coordonnes des cases d'un ensemble
static void JEU_f_priv_list_set(JEU_priv_cases_set_t *ps)
{
  int bs;
  int xe,ye;
  int i;
  char *s;

  //Type d'ensemble?
  switch(ps->t)
  {
    case JEU_SET_LIGNE_E:
      s="Ligne";
      break;
    case JEU_SET_COLONNE_E:
      s="Colonne";
      break;
    case JEU_SET_CARRE_E:
      s="Carre";
      break;
    default:
      s="Erreur";
  }
  ihm_out("%s:", s);

  bs = JEU_v_priv_etat.Boardsize;
  for(i=0; i<bs; i++)
  {
    xe = ps->a_i2x[i];
    ye = ps->a_i2y[i];
    ihm_out("(%d,%d)",xe,ye);
  }
}

//Affiche les valeurs  jouer / prsentes  sur un ensemble de cases.
static void JEU_f_priv_print_set(JEU_priv_cases_set_t *ps)
{
  int n;
  int bs;
  int nb;

  bs = JEU_v_priv_etat.Boardsize;

  //Valeurs vides.
  for(n=1; n<=bs; n++)
  {
    nb = JEU_f_priv_get_set_n2nb(ps,n);
    if (nb > 0)
      ihm_out("%d",n);
  }
  ihm_out("/");
  //Valeurs prsentes.
  for(n=1; n<=bs; n++)
  {
    nb = JEU_f_priv_get_set_n2nb(ps,n);
    if (nb == 0)
      ihm_out("%d",n);
  }
  //Erreurs?
  for(n=1; n<=bs; n++)
  {
    nb = JEU_f_priv_get_set_n2nb(ps,n);
    if (nb < 0)
      ihm_out("!!%d->%d",n,nb);
  }

}

//Ligne de sparation
void JEU_f_priv_drawline(int bs, int es)
{
  int x;
  for(x=0; x<bs; x++) { if (x % es == 0) ihm_out("+"); ihm_out("--"); }
  ihm_out("+\n");
}

//Affiche le nombre de coups jouables sur chaque case.
//Affiche "." si la case est dj occupe.
void JEU_f_print_board(void)
{
  int x, y;
  int nb;
  unsigned char **cases;
  int n;
  char *a_n2e;
  int bs, es;

  bs = JEU_v_priv_etat.plateau.Boardsize;
  es = JEU_v_priv_etat.plateau.Elemsize;

  cases = JEU_v_priv_etat.plateau.cases;
  for(y=0; y<bs; y++)
  {
    //Ligne de sparation?
    if (y % es == 0)
      JEU_f_priv_drawline(bs,es);
    for(x=0; x<bs; x++)
    {
      //Sparateur vertical?
      if (x % es == 0)
        ihm_out("|");
      //Contenu de la case
      n = cases[x][y];
      //Case occupe: afficher "."
      if (n != 0)
        ihm_out(" .");
      //Case libre: compter les coups possibles
      else {
        a_n2e = JEU_v_priv_etat.a_xyn2e[x][y];
        nb = 0;
        for(n=1; n<=bs; n++)
        {
          if(a_n2e[n] == 0) //Valeur absente dans ligne, colonne, carr
            nb++;
        }
        if (nb == JEU_v_priv_etat.a_xy2nb[x][y])
          ihm_out(" %d", nb);
        else
          ihm_out("!%d", nb);
      }
    }
    //Bord droit
    ihm_out("|\n");
  }
  //Bord du bas
  JEU_f_priv_drawline(bs,es);
}

//Pour chaque case:
//  .  =occupe,
//  123=interdit de jouer 'v'
//  X  =on peut jouer 'n'
//  *  =coup forc (seule case d'une ligne,colonne ou carr o 'v' est jouable)
void JEU_f_print_board_n(int v)
{
  int x, y;
  int nb;
  unsigned char **cases;
  int n;
  int bs, es;
  JEU_priv_cases_set_t *psL, *psCL, *psCR; //Ensembles affects.
  int nbL, nbCL, nbCR;

  bs = JEU_v_priv_etat.plateau.Boardsize;
  es = JEU_v_priv_etat.plateau.Elemsize;

  cases = JEU_v_priv_etat.plateau.cases;
  for(y=0; y<bs; y++)
  {
    //Ligne de sparation?
    if (y % es == 0)
      JEU_f_priv_drawline(bs,es);
    for(x=0; x<bs; x++)
    {
      //Sparateur vertical?
      if (x % es == 0)
        ihm_out("|");
      //Contenu de la case
      n = cases[x][y];
      //Case occupe: afficher "."
      if (n != 0)
        ihm_out(" .");
      //Case libre: 'n' est-il interdit?
      else {
        nb = JEU_v_priv_etat.a_xyn2e[x][y][v];
        switch(nb)
        {
          case 0:         //'v' n'est pas interdit ...

            psL = JEU_f_priv_get_set_ligne(y);
            psCL = JEU_f_priv_get_set_col(x);
            psCR = JEU_f_priv_get_set_carre(x,y);
            nbL = JEU_f_priv_get_set_n2nb(psL, v);
            nbCL = JEU_f_priv_get_set_n2nb(psCL, v);
            nbCR = JEU_f_priv_get_set_n2nb(psCR, v);
            if ( (nbL==1) || (nbCL==1) || (nbCR==1) )
              ihm_out(" *"); //'v' est obligatoirement l ...
            else
              ihm_out(" X"); //'v' peut tre l
            break;
          case 1:         //'v' est interdit 1,2 ou 3 fois
          case 2:
          case 3:
            ihm_out(" %d",nb);
            break;
          default:       //Pas d'autre interdiction possible
            ihm_out("!%d",nb);
            break;
        }
      }
    }
    //Bord droit
    ihm_out("|\n");
  }
  //Bord du bas
  JEU_f_priv_drawline(bs,es);
}

//Liste les coups autoriss sur une case.
//Pour chaque valeur, donne le nombre de cases de ligne, colonne, rectangle o on peut le mettre.
void JEU_f_print_list(int x, int y)
{
  char *a_n2e;
  int v;
  int bs;
  JEU_priv_cases_set_t *psL, *psCL, *psCR; //Ensembles affects.
  int nbL, nbCL, nbCR; //Et coups sur ces ensembles
  int nb;

  a_n2e = JEU_v_priv_etat.a_xyn2e[x][y];
  bs = JEU_v_priv_etat.plateau.Boardsize;

  ihm_out("(%d,%d):",x,y);

  //Lister les coups possibles
  nb = 0;
  for(v=1; v<=bs; v++)
  {
    if (a_n2e[v] == 0) //Valeur absente dans ligne, colonne, carr
    {
      ihm_out("%d",v);
      nb++;
    }
  }

  //Vrifier si le compteur est  jour
  if (JEU_v_priv_etat.a_xy2nb[x][y] != nb)
    ihm_out("!(%d)", JEU_v_priv_etat.a_xy2nb[x][y]);

  //Dtails
    psL = JEU_f_priv_get_set_ligne(y);
    psCL = JEU_f_priv_get_set_col(x);
    psCR = JEU_f_priv_get_set_carre(x,y);
    for(v=1; v<=bs; v++)
    {
      nbL = JEU_f_priv_get_set_n2nb(psL, v);
      nbCL = JEU_f_priv_get_set_n2nb(psCL, v);
      nbCR = JEU_f_priv_get_set_n2nb(psCR, v);

      ihm_out("  %d:%d%d%d", v, nbL, nbCL, nbCR);
    }
  ihm_out("\n");

  ihm_out("Ligne  :"); JEU_f_priv_list_set(psL);  JEU_f_priv_print_set(psL);  ihm_out("\n");
  ihm_out("Colonne:"); JEU_f_priv_list_set(psCL); JEU_f_priv_print_set(psCL); ihm_out("\n");
  ihm_out("Carre  :"); JEU_f_priv_list_set(psCR); JEU_f_priv_print_set(psCR); ihm_out("\n");

}


//Traduit la raison d'un coup
void JEU_f_print_expliquer(int idx, JEU_explain_t *pe, JEU_coup_t *pc)
{
  char *s;

  switch(*pe)
  {
    case E_JEU_BALAIE_CARRE:
      s = "Balayage Carr:";
      break;
    case E_JEU_BALAIE_LIGNE:
      s = "Balayage Ligne:";
      break;
    case E_JEU_BALAIE_COLONNE:
      s = "Balayage Colonne:";
      break;
    case E_JEU_UN_COUP_LEGAL:
      s = "Case avec une seule possibilit:";
      break;
    case E_JEU_CASE_SANS_COUP:
      s = "Case sans aucune possibilit:";
      break;
    case E_JEU_ELIM_LIGNE:
      s = "Nettoyage sur une ligne:\n";
      break;
    case E_JEU_ELIM_COLONNE:
      s = "Nettoyage sur une colonne:\n";
      break;
    case E_JEU_TATONNEMENT:
      s = "Essais successifs:";
      break;
    case E_JEU_ELIM_CARRE:
      s = "Nettoyage sur un carr:\n";
      break;
    case E_JEU_ZERO_COUP:
      s = "Aucun coup lgal trouv:\n";
      break;
    default:
      s = "Inconnu\n";
  }

  switch(pc->t)
  {
    case E_JEU_PRIV_COUP_CATEGORY_WRITE:
      if ((*pe != JEU_v_priv_last_cause) || (JEU_v_priv_last_idx >= idx) )
      {
        ihm_out("\n  %d: %s", idx, s);
      }
      {
        char c;
        c = (char)(pc->s_write.x + ((int) 'A'));
        ihm_out("%c%d=%d; ", c, pc->s_write.y, pc->s_write.n);
      }
      break;
    case E_JEU_PRIV_COUP_CATEGORY_CUT:
      if (*pe != JEU_v_priv_last_cause)
        ihm_out("\n  %d: %s    ", idx, s);
      else
        ihm_out("    ", idx);
      JEU_f_priv_print_subset(&JEU_v_priv_etat.a_sets[pc->s_cut.i_set],
                              pc->s_cut.a_subset, pc->s_cut.a_vals);
      break;
    default:
      ihm_out("Cause inconnue ... je ne sais que dire!\n");
  }

  JEU_v_priv_last_cause = *pe;
  JEU_v_priv_last_idx = idx;
}


// ---- Jeu: routines prives -----

#define DEBUG_ON 0

//Compte les cases inoccupes dans un ensemble.
static int JEU_f_priv_get_set_free(JEU_priv_cases_set_t *ps)
{
  int i, x, y;
  int bs;
  int nb;

  bs = JEU_v_priv_etat.Boardsize;
  nb = 0;              //Compter ...
  for(i=0; i<bs; i++)  //...les cases de l'ensemble...
  {
    x = ps->a_i2x[i];
    y = ps->a_i2y[i];
    if ( JEU_v_priv_etat.plateau.cases[x][y] == 0 ) //...qui sont libres...
    {
      nb++;
    }
  }
  return nb;
}

//Renvoie le nombre de cases LIBRES o 'n' est un coup lgal
static int JEU_f_priv_get_set_n2nb(JEU_priv_cases_set_t *ps, int n)
{
  int i, x, y;
  int bs;
  int nb;

  bs = JEU_v_priv_etat.Boardsize;
  nb = 0;              //Compter ...
  for(i=0; i<bs; i++)  //...les cases de l'ensemble...
  {
    x = ps->a_i2x[i];
    y = ps->a_i2y[i];
    if ( JEU_v_priv_etat.plateau.cases[x][y] == 0 ) //...qui sont libres...
    {
      if (JEU_v_priv_etat.a_xyn2e[x][y][n] == 0)       //...et o on peut jouer n...
      {
        nb++;
        IFDBG { ihm_out("%d jouable en (%d,%d)\n", n, x, y); }
      }
    }
  }
  return nb;
}

#define DEBUG_ON 0

//Pointeurs vers les ensembles colonnes
static JEU_priv_cases_set_t* JEU_f_priv_get_set_col(int x)
{
  int bs;
  bs = JEU_v_priv_etat.Boardsize;
  return &JEU_v_priv_etat.a_sets[bs+bs+x];
}

//Idem pour les lignes
static JEU_priv_cases_set_t* JEU_f_priv_get_set_ligne(int y)
{
  int bs;

  bs = JEU_v_priv_etat.Boardsize;
  return &JEU_v_priv_etat.a_sets[bs+y];
}

//Idem pour les colonnes
static JEU_priv_cases_set_t* JEU_f_priv_get_set_carre(int x, int y)
{
  int a, b, i;
  //int bs;
  int es;

  //bs = JEU_v_priv_etat.Boardsize;
  es = JEU_v_priv_etat.Elemsize;

  a = x / es;
  b = y / es;
  i = a+(es*b);

  return &JEU_v_priv_etat.a_sets[i];
}

//Initialise un ensemble par dfaut,
//  en omettant juste la liste des coordonnes des cases.
static void JEU_f_priv_default_mem_set(JEU_priv_cases_set_t *ps)
{
  int n;
  int bs;

  bs = JEU_v_priv_etat.Boardsize;

  //Rservation de mmoire
  ps->a_i2x  = (int*) m_dbg_malloc(bs*sizeof(int));
  ps->a_i2y  = (int*) m_dbg_malloc(bs*sizeof(int));
  //ps->a_n2nb = (int*) m_dbg_alloc((bs+1)*sizeof(int));
  ps->a_n2e  = (int*) m_dbg_malloc((bs+1)*sizeof(int));

  //Coups possibles et cases libres par dfaut.
  for(n=1; n<=bs; n++)
  {
    //ps->a_n2nb[n] = bs; //'n' est jouable partout.
    ps->a_n2e[n]  = 0;  //'n' n'est prsent nulle part.
  }
}

#define DEBUG_ON 0

//Rservations de mmoire et coordonnes pour un ensemble
//Colonne: mmoire
static int JEU_f_priv_mem_set_col(int x)
{
  JEU_priv_cases_set_t *ps;
  int bs;
  int y;

  ps = JEU_f_priv_get_set_col(x);
  bs = JEU_v_priv_etat.Boardsize;

  JEU_f_priv_default_mem_set(ps);

  ps->t = JEU_SET_COLONNE_E;
  for (y=0; y<bs; y++)
  {
    ps->a_i2x[y] = x;
    ps->a_i2y[y] = y;
  }

  IFDBG { JEU_f_priv_list_set(ps); }

  return 0;
}

//Ligne: mmoire
static int JEU_f_priv_mem_set_ligne(int y)
{
  JEU_priv_cases_set_t *ps;
  int bs;
  int x;

  ps = JEU_f_priv_get_set_ligne(y);
  bs = JEU_v_priv_etat.Boardsize;

  JEU_f_priv_default_mem_set(ps);

  ps->t = JEU_SET_LIGNE_E;
  for (x=0; x<bs; x++)
  {
    ps->a_i2x[x] = x;
    ps->a_i2y[x] = y;
  }

  IFDBG { JEU_f_priv_list_set(ps); }

  return 0;
}

//Carr: mmoire
static int JEU_f_priv_mem_set_carre(int x, int y)
{
  JEU_priv_cases_set_t *ps;
  int es; //Abbrviation 'Elemsize'
  int x0, y0;
  int i, a, b; //Intermdiaires

  ps = JEU_f_priv_get_set_carre(x,y);
  es = JEU_v_priv_etat.Elemsize;

  JEU_f_priv_default_mem_set(ps);

  ps->t = JEU_SET_CARRE_E;

  x0=(x/es) * es;
  y0=(y/es) * es;

  i=0;
  for(b=0; b<es; b++)
  {
    for(a=0; a<es; a++)
    {
      ps->a_i2x[i] = x0+a;
      ps->a_i2y[i] = y0+b;
      i+=1;
    }
  }

  IFDBG { JEU_f_priv_list_set(ps); }

  return 0;
}
#define DEBUG_ON 0

//Rend la mmoire rserve pour un ensemble
static void JEU_f_priv_mem_set_free(JEU_priv_cases_set_t *ps)
{
  free(ps->a_i2x);
  free(ps->a_i2y);
}

#define DEBUG_ON 0

//Compte les valeurs dj prsentes des coups jouables dans *ps
//Initialise les champs a_xyn2e[x,y,v] (du jeu) et a_n2e[v] (de l'ensemble)
//      * Les champs a_xyn2e[x,y,v] seront cohrents aprs le balayage de TOUS les ensembles.
//Au passage, produit un message d'erreur si une valeur dfendue est l.
//Arguments:
//  ps : pointeur sur un descripteur d'ensemble.
//Retour:
//  TRUE  : initialisation sans problme
//  FALSE : contenu incorrect (valeur rpte ou mauvaise valeur)
static bool_t JEU_f_priv_init_part_set(JEU_priv_cases_set_t *ps)
{
  int bs;
  char *p_n2e;  // Liste des coups permis sur la case courante
  int xe, ye;  // Coordonnes auxiliaires
  int i;
  char v;    // Index auxiliaires

  bs = JEU_v_priv_etat.Boardsize;

  //Recenser les valeurs qui sont dj l.
  for(v=1;v<=bs;v++)
    ps->a_n2e[v] = 0;     //v absent de l'ensemble
  for(i=0; i<bs; i++)
  {
    xe = ps->a_i2x[i];
    ye = ps->a_i2y[i];
    v  = JEU_v_priv_etat.plateau.cases[xe][ye];
    //Valeur correcte?
    if ((v<0) || (v>bs))
    {
      ihm_out("ERREUR: valeur d'entre %d incorrecte en (%d,%d)\n", v, xe, ye);
      return FALSE;
    }
    if (v != 0)
    {
      ps->a_n2e[v] +=1;  //v : occurrence dans l'ensemble.
      //Rptition?
      if (ps->a_n2e[v] > 1)
      {
        ihm_out("ERREUR: valeur %d rpte en (%d,%d)\n", v, xe, ye);
        return FALSE;
      }
    }
  }

  //Reporter a dans les cases
  for(v=1;v<=bs;v++)
  {
    if (ps->a_n2e[v] > 0)     //Si la valeur est prsente dans l'ensemble
    {
      IFDBG { ihm_out("Report de %d dans l'ensemble...\n",v); }
      for(i=0; i<bs; i++)   //Pour chaque case de l'ensemble
      {
        xe = ps->a_i2x[i];
        ye = ps->a_i2y[i];
        p_n2e = JEU_v_priv_etat.a_xyn2e[xe][ye];
        p_n2e[v] += (char) 1;      //Compter une occurrence de la valeur autour de cette case
      } //for(i=...)
    }
  }     //for(v=...)

  //TERMINAISON
  IFDBG{ ihm_out("-----------------------------------\n"); }

  return TRUE;
}

#define DEBUG_ON 0

//Consquences d'un coup sur un ensemble de cases
//NOTE: La variable JEU_v_priv_etat ne sera  nouveau cohrente
//  qu'aprs l'excution de cette fonction pour TOUS les
//  ensembles concerns par le coup.
//NOTE: 'ecrire' fait actualiser, le cas chant, le nombre de cases occupes de l'ensemble.
#define DEBUG_ON 0
static void JEU_f_priv_put_n_set(JEU_priv_cases_set_t *ps, char v, bool_t ecrire)
{
  int bs;
  int i;
  char val;

  bs = JEU_v_priv_etat.Boardsize;

  if (ecrire != FALSE)
    ps->a_n2e[v]  +=1;   //... et 'v' y apparat une fois (de plus).

  //Cases o on ne peut plus jouer v
  for(i=0; i<bs; i++)  //Pour chaque case de l'ensemble
  {
    int x,y;

    x = ps->a_i2x[i];
    y = ps->a_i2y[i];
    val = (char) (JEU_v_priv_etat.a_xyn2e[x][y][v] + 1); //Et un empchement de plus pour cette valeur.
    JEU_v_priv_etat.a_xyn2e[x][y][v] = val;
    if (val == 1)                             //Si 'v' pouvait tre jou l ...
    {
      JEU_v_priv_etat.a_xy2nb[x][y] -= (char) 1;     //...eh bien ce n'est plus le cas.
    }
  }
}

#define DEBUG_ON 0

//Consquences de l'annulation d'un coup sur un ensemble de cases
//NOTE: La variable JEU_v_priv_etat ne sera  nouveau cohrente
//  qu'aprs l'excution de cette fonction pour TOUS les
//  ensembles concerns par le coup.
//NOTE: 'ecrire' fait actualiset, le cas chant, le nombre de cases occupes de l'ensemble.
static void JEU_f_priv_del_n_set(JEU_priv_cases_set_t *ps, char v, bool_t ecrire)
{
  int i;
  int bs;
  char val;

  bs = JEU_v_priv_etat.Boardsize;

  //Cases o v redevient jouable
  for(i=0; i<bs; i++)  //Pour chaque case de l'ensemble
  {
    int x,y;

    x = ps->a_i2x[i];
    y = ps->a_i2y[i];
    val = (char) (JEU_v_priv_etat.a_xyn2e[x][y][v] - 1); //Et un empchement de moins pour cette valeur.
    JEU_v_priv_etat.a_xyn2e[x][y][v] = val;
    if (val == 0)                             //S'il redevient possible de jouer 'v'  cet endroit ...
    {
      JEU_v_priv_etat.a_xy2nb[x][y] +=  (char) 1; //S'en souvenir pour la case ...
    }
  }

  if (FALSE != ecrire)
    ps->a_n2e[v]  -=1;   //... et 'v' y apparat une fois de moins.
}

#define DEBUG_ON 0

//Recherche rcursive d'un sous-ensemble de 'nb_subset' cases libres de *ps,
//  demandant exactement 'nb_subset' valeurs possibles.
//Arguments:
//  *ps       : l'ensemble o on prend les cases
//  nb_subset : cardinal du sous-ensemble que l'on veut trouver
//  nb        : nombre de cases dj trouves.
//  idx0      : index de *ps o commencer la recherche des autres cases
//  a_subset[0..(bs-1)] : =0 pour les index hors-ensemble, =1 pour ceux de l'ensemble [jusqu' prsent].
//  a_vals[0..bs] : nombre de fois o chaque valeur est possible dans une case de a_subset.
//Rsultat:
//  TRUE si ensemble trouv, FALSE sinon.
//  Si TRUE, les index correspondants de a_subset[] et de a_vals[] sont mis  1
static bool_t JEU_f_priv_find_subset(JEU_priv_cases_set_t *ps, int nb_subset, int nb, int idx0, char *a_subset, char *a_vals)
{
  int bs;
  int i, x, y, v; //Indices de boucle.
  int nb_vals;

  bs = JEU_v_priv_etat.Boardsize;
  //Si la taille dsire n'est pas encore atteinte
  if (nb_subset > nb)
  {
    for(i=idx0; i<bs; i++)
    {
      x = ps->a_i2x[i];
      y = ps->a_i2y[i];
      //Si la case est libre ...
      if (JEU_v_priv_etat.plateau.cases[x][y] != 0)
        continue;

      //Report dans a_subset[]
      a_subset[i]=1;
      //Report dans a_vals[];
      for(v=1; v<=bs; v++)
      {
        if (JEU_v_priv_etat.a_xyn2e[x][y][v] == 0)
        {
          a_vals[v] += (char) 1;
        }
      }

      //Nombre d'lments de a_vals[] ?
      nb_vals = 0;
      for(v=1; v<=bs; v++)
      {
        if (a_vals[v] != 0)
          nb_vals++;
      }

      //Intrt de poursuivre?
      if (nb_vals <= nb_subset)
      {
        //Si a vaut le coup de poursuivre...
        if (TRUE == JEU_f_priv_find_subset(ps, nb_subset, nb+1, i+1, a_subset, a_vals))
          return TRUE;
      }

      //Si c'est rat: retirer de a_vals[]
      for(v=1; v<=bs; v++)
      {
        if (JEU_v_priv_etat.a_xyn2e[x][y][v] == 0)
        {
          a_vals[v] -= (char) 1;
        }
      }
      //Retirer de a_subset[]
      a_subset[i]=0;
    } //for(...)
    return FALSE;
  }
  else
  //Si la taille dsire est atteinte,
  //Permet-elle un lagage des possibilits?
  {
    //NB: ici la condition de l'appel rcursif impose
    //que a_subset[] mentionne 'nb_subset' indices, et que a_vals[] n'en cite pas plus.
    for(i=0; i<bs; i++)
    {
      x = ps->a_i2x[i];
      y = ps->a_i2y[i];
      //Si la case est libre ...
      if (JEU_v_priv_etat.plateau.cases[x][y] != 0)
        continue;
      //Si elle n'est pas mentionne par a_subset[]
      if (a_subset[i] != 0)
        continue;
      //S'il y a au moins une valeur ...
      for(v=1; v<=bs; v++)
      {
        if (a_vals[v] > 0) //...mentionne par a_vals[]...
        {
          if (JEU_v_priv_etat.a_xyn2e[x][y][v] == 0) //...et jouable ici...
          {
            IFDBG { ihm_out("Elagage de %d en (%d,%d)\n", v, x, y); }
            return TRUE;                             //Alors c'est bon
          }
        }
      }
    }
    //Ensemble ok, mais sans possibilit d'lagage.
    return FALSE;
  }
  IFDBG { ihm_out("JEU_f_priv_find_subset: interdit!\n"); }
  return FALSE;
}

#define DEBUG_ON 0

// ---- Jeu: routines publiques -----

//Explication par dfaut (aucune)
void JEU_f_init_explain(JEU_explain_t *pe)
{
  *pe = E_JEU_ZERO_COUP;
}

//Initialise le jeu  partir de la position de dpart donne.
//Renvoie FALSE si cette position de dpart est invalide.
//Sinon renvoie TRUE
bool_t JEU_f_init(JEU_position_t *p_position)
{
  int bs, es; //Abbrger les dimensions
  int x, y;
  JEU_position_t *ppos; //Abrger &JEU_v_priv_etat.plateau
  int v;                //Valeurs jouables (ou non)
  JEU_priv_cases_set_t *ps; //Pointe un ensemble de cases
  int i;
  bool_t r;             //Contrle de validit.

  //Rfrences de base
  bs = p_position->Boardsize;
  es = p_position->Elemsize;
  ppos = &JEU_v_priv_etat.plateau;

  //Pas encore de coup  expliquer
  JEU_v_priv_last_cause = E_JEU_UNDEFINED;
  JEU_v_priv_last_idx   = 0;

  //La grille a l'air correcte par dfaut;
  r = TRUE;

  //Recopier la position
  ppos->cases = m_dbg_malloc2(unsigned char,bs,bs); //Tableau 2D: bs*bs
  for(x=0; x<bs; x++)
  {
    for(y=0; y<bs; y++)
      ppos->cases[x][y] = p_position->cases[x][y];
  }
  JEU_v_priv_etat.plateau.Boardsize = bs; //Champs originaux
  JEU_v_priv_etat.plateau.Elemsize = es;
  JEU_v_priv_etat.Boardsize = bs;           //Champs recopis
  JEU_v_priv_etat.Elemsize = es;

  //Rserver de la mmoire pour les ensembles
  JEU_v_priv_etat.a_sets = (JEU_priv_cases_set_t*) m_dbg_malloc(3*bs*sizeof(JEU_priv_cases_set_t));
  for(x=0;x<bs;x++)
    JEU_f_priv_mem_set_col(x);
  for(y=0;y<bs;y++)
    JEU_f_priv_mem_set_ligne(y);
  for(x=0;x<bs;x+=es)
  {
    for(y=0;y<bs;y+=es)
      JEU_f_priv_mem_set_carre(x,y);
  }

  //Rservation mmoire pour les coups autoriss
  JEU_v_priv_etat.a_xyn2e = m_dbg_malloc3(char, bs, bs, bs+1); //Tableau 3D:bs*bs*(bs+1)
  for(x=0;x<bs;x++)
  {
    for(y=0;y<bs;y++)
    {
      for(v=1;v<=bs+1;v++)
      {
        JEU_v_priv_etat.a_xyn2e[x][y][v] = 0; //Jusqu' preuve du contraire, v est absent de ligne, colonne, carr de (x,y)
      }
    }
  }

  //Compter les valeurs dj prsentes
  for(i=0;i<3*bs;i++) //Dans les colonnes, lignes, carrs
  {
    bool_t b;
    ps = &JEU_v_priv_etat.a_sets[i];
    b = JEU_f_priv_init_part_set(ps); //Comptages sur l'ensemble
    r = r && b;                       //Validit du remplissage de l'ensemble
  }
  //Ici le tableau JEU_v_priv_etat.a_xyn2e[...] est correctement rempli

  //Compter les coups lgaux pour chaque case.
  JEU_v_priv_etat.a_xy2nb = m_dbg_malloc2(char,bs,bs);
  for(x=0; x<bs; x++)
  {
    for(y=0; y<bs; y++)
    {
      char n;
      char *n2e;

      n = 0;
      n2e = JEU_v_priv_etat.a_xyn2e[x][y];
      for(v=1; v<=bs; v++) //Compter les coups lgaux en (x,y).
      {
        if (n2e[v] == 0)
          n++;
      }
      JEU_v_priv_etat.a_xy2nb[x][y] = n;
    }
  }

  //Pas de problme  l'initialisation
  return r;
}

////Nombre d'tapes d'exploration  prvoir
////REQUIS: JEU_f_init() appel auparavant.
//int JEU_f_get_explore_size(void)
//{
//  int  bs;
//  bs = JEU_v_priv_etat.Boardsize;
//
//  return (bs*bs); //On pourrait se limiter au nombre des cases vides.
//}

#define DEBUG_ON 0

//Recherche les coups intressants  jouer et renvoie le premier trouv ... s'il existe
//NOTE: cette liste est souvent plus courte que la liste des coups lgaux
//Renvoie TRUE si au moins un coup est trouv, FALSE si une impasse est dtecte.
bool_t JEU_f_lister_coups(ALG_list_coups_t *pcL)
{
  bool_t trouve;
  int i;    //Index
  int x,y;  //Coordonnes de case
  int v;    //Valeur  jouer
  int ic;   //Index de case
  JEU_coup_t coup;
  int bs;
  JEU_priv_cases_set_t *ps;
  int nb;
  int xmin, ymin; //Tester coups uniques et impasses
  int nb_min;

  bs = JEU_v_priv_etat.Boardsize;

  //Vider la liste des coups
  ALG_f_clear_list(pcL);

  trouve = FALSE;

  //Vrifier que l'on n'est pas dans une impasse
  nb_min = bs+1;
  for(x=0; x<bs; x++) //Pour chaque case ...
  {
    for(y=0; y<bs; y++)
    {
      if (JEU_v_priv_etat.plateau.cases[x][y] == 0) // ... libre ...
      {
        nb = JEU_v_priv_etat.a_xy2nb[x][y];         // ... compter les coups lgaux ...
        if (nb < nb_min)                            // ... conserver le minimum ...
        {
          nb_min = nb;
          xmin = x;
          ymin = y;
        }
      }
    }
  }

  if (IHM_v_opt_impasse)
  {
    if (nb_min == 0) //S'il y a une case o on ne peut pas jouer ...
    {
     pcL->explain = E_JEU_CASE_SANS_COUP;
     IFDBG { ihm_out("Impasse en (%d,%d)\n", xmin, ymin); }
     return FALSE;
    }
  }

  //Chercher en balayant par carr/ligne/colonne
  if (IHM_v_opt_balai)
  {
    for(i=0; i<3*bs; i++) //Pour chaque ensemble
    {
      ps = &JEU_v_priv_etat.a_sets[i];
      for(v=1; v<=bs; v++)     //Pour chaque valeur ...
      {
        nb = JEU_f_priv_get_set_n2nb(ps, v);
        if (nb == 1)  //Y a-t-il une unique case de l'ensemble o la caser?
        {
          for(ic=0; ic<bs; ic++) //Chercher de quelle case il s'agit.
          {
            x = ps->a_i2x[ic];
            y = ps->a_i2y[ic];
            if (JEU_v_priv_etat.plateau.cases[x][y] == 0) //Elle doit tre libre
            {
              if (JEU_v_priv_etat.a_xyn2e[x][y][v] == 0)  //...Et le coup ne doit tre interdit par rien.
              {
                coup.t = E_JEU_PRIV_COUP_CATEGORY_WRITE;
                coup.s_write.x = x;
                coup.s_write.y = y;
                coup.s_write.n = v;
                switch(ps->t)
                {
                  case JEU_SET_LIGNE_E:
                    pcL->explain = E_JEU_BALAIE_LIGNE;
                    break;
                  case JEU_SET_COLONNE_E:
                    pcL->explain = E_JEU_BALAIE_COLONNE;
                    break;
                  case JEU_SET_CARRE_E:
                    pcL->explain = E_JEU_BALAIE_CARRE;
                    break;
                 }

                IFDBG { ihm_out("Coup forc: (%d,%d)=%d\n",x,y,v); }

                ALG_f_add_list(pcL, &coup);

                trouve = TRUE; //Pas la peine d'en chercher un autre.
                goto L_case_unique;
              }
            }
          }  //Fin de recherche de la case
        }
      }      //Fin de recherche de la valeur
    }        //Fin de parcours des ensembles.
  }

  //Chercher les cases avec un seul coup lgal
  L_case_unique:
  if (IHM_v_opt_unique)
  {
    if (trouve == FALSE)
    {
      //Chercher parmi les cases ...
      for(x=0; x<bs; x++)
      {
        for(y=0; y<bs; y++)
        {
          if (JEU_v_priv_etat.plateau.cases[x][y] == 0) // ... qui sont libres
          {
            if (JEU_v_priv_etat.a_xy2nb[x][y] == 1)     // ... et o il n'y a qu'un coup  jouer
            {
              //IFDBG { ihm_out("L_case_unique:B: (%d,%d) ????", x, y);}
              for(v=1;v<=bs;v++)
              {
                if (JEU_v_priv_etat.a_xyn2e[x][y][v] == 0) // ... quel est le coup en question
                {
                  coup.t = E_JEU_PRIV_COUP_CATEGORY_WRITE;
                  coup.s_write.x = x;
                  coup.s_write.y = y;
                  coup.s_write.n = v;
                  pcL->explain = E_JEU_UN_COUP_LEGAL;
                  IFDBG { ihm_out("Un coup lgal: (%d,%d)=%d\n",x,y,v); }
    
                  ALG_f_add_list(pcL, &coup);
                  trouve = TRUE;        //Pas la peine d'en chercher un autre.
                  goto L_sous_ensemble;
                }
              }
            }
          }
        }
      }
    }
  }

  //Chercher les liminations par sous-ensemble
  L_sous_ensemble:
  if (IHM_v_opt_nettoyer)
  {
    if (trouve == FALSE)
    {
      char *a_subset, *a_vals;
      int nb_subset, nb_free_cells;

      a_subset = (char*) m_dbg_malloc(bs*sizeof(char));
      a_vals = (char*) m_dbg_malloc((bs+1)*sizeof(char));

      //Chercher le plus petit sous-ensemble possible
      for(nb_subset=2; nb_subset<bs; nb_subset++)
      {
        //Au sein de chaque ensemble...
        for(i=0; i<3*bs; i++)
        {
          //Combien de cases libres?
          ps = &JEU_v_priv_etat.a_sets[i];
          nb_free_cells = JEU_f_priv_get_set_free(ps);

          //Il en faut plus que la taille du sous-ensemble recherch
          if (nb_free_cells <= nb_subset)
            continue;

          for(x=0; x<bs; x++)
          {
            a_subset[x] = 0;
            a_vals[x] = 0;
          }

          //Recherche rcursive d'un ss-ensemble de 'nb_subset' cases libres
          //attendant exactement 'nb_subset' valeurs possibles,
          //dont au moins une est 'lgale' en-dehors de ce sous-ensemble.
          trouve = JEU_f_priv_find_subset(ps, nb_subset, 0, 0, a_subset, a_vals);
          if (trouve == TRUE)
          {
            IFDBG { ihm_out("Ensemble trouv\n"); }
            //Remplir une description de coup ad-hoc
            coup.t = E_JEU_PRIV_COUP_CATEGORY_CUT;
            coup.s_cut.i_set = i;           //Index de l'ensemble parent
            coup.s_cut.a_subset = a_subset; //[]=1 pour les lments prsents
            coup.s_cut.a_vals   = a_vals;   //[]!=0 pour les valeurs prsentes
            switch(ps->t)
            {
              case JEU_SET_LIGNE_E:
                pcL->explain = E_JEU_ELIM_LIGNE;
                break;
              case JEU_SET_COLONNE_E:
                pcL->explain = E_JEU_ELIM_COLONNE;
                break;
              case JEU_SET_CARRE_E:
                pcL->explain = E_JEU_ELIM_CARRE;
                break;
              default:
                ihm_out("ERREUR!\n");
            }

            ALG_f_add_list(pcL, &coup);     //Mettre a dans la liste des coups trouvs

            IFDBG { JEU_f_priv_print_subset(ps, a_subset, a_vals); }

            goto L_tatonner; //NB: sans librer a_subset ni a_vals, dsormais possds par 'coup'
          }
        }
      }

      //Si rien n'est trouv, librer 'a_subset' et 'a_vals' qui ne servent plus  rien.
      m_dbg_free(a_subset);
      m_dbg_free(a_vals);
    }
  }

  //Si rien de tout a ne marche, ttonner par essais successifs:
  //  Reprer la case qui a le moins de coups lgaux.
  //  Et dcider d'examiner ceux-cis.
  L_tatonner:
  if (trouve == FALSE)
  {
    //Reprer les valeurs jouables en (xmin, ymin)
    if (nb_min <= bs) //S'il y avait au moins une case libre (cf. plus haut)
    {
      for(v=1; v<=bs; v++)
      {
        if (JEU_v_priv_etat.a_xyn2e[xmin][ymin][v] == 0)
        {
          IFDBG { ihm_out("Tatonnement: (%d,%d) = %d\n", xmin, ymin, v); }
          coup.t = E_JEU_PRIV_COUP_CATEGORY_WRITE;
          coup.s_write.x = xmin;
          coup.s_write.y = ymin;
          coup.s_write.n = v;
          pcL->explain = E_JEU_TATONNEMENT;

          ALG_f_add_list(pcL, &coup);
        }
      }
    }
    else
    {
      pcL->explain = E_JEU_ZERO_COUP; //Toutes les cases sont occupes?
    }
  }
  return TRUE;
}

#define DEBUG_ON 0

// Regarde si on a atteint une solution
bool_t JEU_f_est_solution(void)
{
  int x,y;
  int bs;
  bs = JEU_v_priv_etat.Boardsize;

  //S'il y a au moins une case vide ...
  for(x=0; x<bs; x++)
  {
    for(y=0; y<bs; y++)
    {
      if (JEU_v_priv_etat.plateau.cases[x][y] == 0)
        return FALSE;
    }
  }
  //Si toutes les cases ont t remplies.
  return TRUE;
}

#define DEBUG_ON 1

//Jouer un coup
void JEU_f_jouer(JEU_coup_t *pc)
{
  JEU_priv_cases_set_t *ps; //Ensembles affects.
  int x,y;
  char v;
  int i;  //Indice de boucle sur un ensemble.
  int bs;

  switch(pc->t)
  {

    //Ecrire un nombre dans une case
    case E_JEU_PRIV_COUP_CATEGORY_WRITE:
      x = pc->s_write.x;
      y = pc->s_write.y;
      v = (char) pc->s_write.n;

      //Modifier l'tat de la ligne
      ps = JEU_f_priv_get_set_ligne(y);
      JEU_f_priv_put_n_set(ps, v, TRUE);

      //Modifier l'tat de la colonne
      ps = JEU_f_priv_get_set_col(x);
      JEU_f_priv_put_n_set(ps, v, TRUE);

      //Modifier l'tat du carr
      ps = JEU_f_priv_get_set_carre(x,y);
      JEU_f_priv_put_n_set(ps, v, TRUE);

      //Modifier l'tat de la case
      JEU_v_priv_etat.plateau.cases[x][y] = v;
      break;

    //Elaguer la liste des coups jouables
    case E_JEU_PRIV_COUP_CATEGORY_CUT:

      bs = JEU_v_priv_etat.Boardsize;
      ps = &JEU_v_priv_etat.a_sets[pc->s_cut.i_set];
      if (IHM_v_opt_expliquer)
      	ihm_out("  ==>valeurs interdites en");
      for(i=0; i<bs; i++)
      {
        if (pc->s_cut.a_subset[i] == 0)
        {
          //IFDBG { ihm_out("Subset A"); }
          //Si la case n'est pas dans l'ensemble
          x = ps->a_i2x[i];
          y = ps->a_i2y[i];
          //... et si elle est libre ...
          if (JEU_v_priv_etat.plateau.cases[x][y] == 0)
          {
            bool_t b_interdit = FALSE;
            for(v=1; v<=bs; v++)
            {
              JEU_priv_cases_set_t *psL, *psCL, *psCR;

              //Et si rien ne dfend de mettre 'v'  cet endroit
              psL = JEU_f_priv_get_set_ligne(y);
              psCL = JEU_f_priv_get_set_col(x);
              psCR = JEU_f_priv_get_set_carre(x,y);

              if ((pc->s_cut.a_vals[v] != 0) && (psL->a_n2e[v]==0) && (psCL->a_n2e[v]==0) && (psCR->a_n2e[v]==0))  //...et si elle est  laguer...
              {
                int val;
                val = JEU_v_priv_etat.a_xyn2e[x][y][v] + 1; //Compter une interdiction pour cause de balayage.
                JEU_v_priv_etat.a_xyn2e[x][y][v] = (char) val;
                if (val == 1)
                {
                  b_interdit = TRUE;                          //Au moins une valeur  interdire en (x,y)!
                  JEU_v_priv_etat.a_xy2nb[x][y] -= (char) 1;  //Et un coup lgal en moins si ncessaire
                }
              }
            }
            if ((b_interdit == TRUE) && IHM_v_opt_expliquer) //Une valeur  interdire sur cette case?
              ihm_out(" %c%d;", (char)(x + ((int) 'A')), y);
          }
        }
      }
      if (IHM_v_opt_expliquer)
        ihm_out("\n");
      break;

    default:
      break;
   }
}

#define DEBUG_ON 0

//Revenir sur un coup
void JEU_f_annuler(JEU_coup_t *pc)
{
  JEU_priv_cases_set_t *ps; //Ensembles affects.
  int x,y;
  char v, i; // Indices de boucle
  int bs;

   switch(pc->t)
   {

     case E_JEU_PRIV_COUP_CATEGORY_WRITE:
       x = pc->s_write.x;
       y = pc->s_write.y;
       v = (char) pc->s_write.n;

       //Modifier l'tat de la case
       JEU_v_priv_etat.plateau.cases[x][y] = 0;

       //Modifier l'tat de la ligne
       ps = JEU_f_priv_get_set_ligne(y);
       JEU_f_priv_del_n_set(ps, v, TRUE);

       //Modifier l'tat de la colonne
       ps = JEU_f_priv_get_set_col(x);
       JEU_f_priv_del_n_set(ps, v, TRUE);

       //Modifier l'tat du carr
       ps = JEU_f_priv_get_set_carre(x,y);
       JEU_f_priv_del_n_set(ps, v, TRUE);
       break;

     //Rtablir des coups jouables
     case E_JEU_PRIV_COUP_CATEGORY_CUT:
       ps = &JEU_v_priv_etat.a_sets[pc->s_cut.i_set];
             bs = JEU_v_priv_etat.Boardsize;
       for(i=0; i<bs; i++)
       {
         if (pc->s_cut.a_subset[i] == 0)
         {
           //Si la case n'est pas dans l'ensemble
           x = ps->a_i2x[i];
           y = ps->a_i2y[i];
           //... et si elle est libre ...
           if (JEU_v_priv_etat.plateau.cases[x][y] == 0)
           {
             for(v=1; v<=bs; v++)
             {
               JEU_priv_cases_set_t *psL, *psCL, *psCR;

               //Et si rien ne dfend de remettre 'v'  cet endroit
               psL = JEU_f_priv_get_set_ligne(y);
               psCL = JEU_f_priv_get_set_col(x);
               psCR = JEU_f_priv_get_set_carre(x,y);

               if ((pc->s_cut.a_vals[v] != 0) && (psL->a_n2e[v]==0) && (psCL->a_n2e[v]==0) && (psCR->a_n2e[v]==0))  //...et si l'interdiction vient d'un balayage ...
               {
                 int val;
                 val = (char) (JEU_v_priv_etat.a_xyn2e[x][y][v] - 1); //Alors compter une interdiction de moins ...
                 JEU_v_priv_etat.a_xyn2e[x][y][v] = (char) val;
                 if (val == 0)
                   JEU_v_priv_etat.a_xy2nb[x][y]  += (char) 1; //Et, le cas chant, un coup lgal de plus.
               }
             }
           }
         }
       }
       break;

     default:
       break;
  }  //case
}

//Renvoie un pointeur sur un descripteur de position
//NB: ne SURTOUT pas crire dans ce pointeur
JEU_position_t *JEU_f_get_position(void)
{
  return &JEU_v_priv_etat.plateau;
}

//Libre la mmoire qu'un coup peut ventuellement
//avoir rserve
void JEU_f_clear_coup(JEU_coup_t *pc)
{
  if (pc->t == E_JEU_PRIV_COUP_CATEGORY_CUT)
  {
    m_dbg_free(pc->s_cut.a_subset);
    m_dbg_free(pc->s_cut.a_vals);
  }
}

//Libre toute la mmoire que le module 'JEU' avait alloue.
void JEU_f_reset(void)
{
  int bs; //Abbrger les dimensions
  JEU_position_t *ppos; //Abrger &JEU_v_priv_etat.plateau
  int i;

  //Rfrences de base
  bs = JEU_v_priv_etat.Boardsize;
  ppos = &JEU_v_priv_etat.plateau;

  //Librer la copie de la position
  m_dbg_mfree2(ppos->cases,bs,bs); //Tableau 2D: bs*bs

  //Librer la mmoire pour les ensembles
  for(i=0;i<3*bs; i++)
    JEU_f_priv_mem_set_free(&JEU_v_priv_etat.a_sets[i]);
  m_dbg_free(JEU_v_priv_etat.a_sets);

  //Librer la mmoire des coups autoriss
  m_dbg_mfree3(JEU_v_priv_etat.a_xyn2e, bs, bs, bs+1); //Tableau 3D:bs*bs*(bs+1)

  //Librer les compteurs des coups lgaux pour chaque case.
  m_dbg_mfree2(JEU_v_priv_etat.a_xy2nb, bs, bs);
}

// ===== FIN DE FICHIER =====
